# Script that automatically creates different versions of a routine with different precisions

# Import classes and functions from UtilsRPE
from AutoRPE.UtilsRPE.CurrentBlock import CurrentBlock
from AutoRPE.UtilsRPE.Inserter import replace_real_declaration
from AutoRPE.UtilsRPE.SourceManager import preprocess_sources, parse_sources, load_vault
from AutoRPE.UtilsRPE.Cleaner import remove_comments

from os import mkdir
from os.path import join, isdir
from shutil import rmtree
import re
import itertools


def get_command_line_arguments():
    from argparse import ArgumentParser

    help_text = """
    This script takes as arguments the input source (-s/--input-source)
    and the name of a routine (-r/--routine) and creates different versions
    of this routine accouting for all different precision combinations that can happen.
    An additional argument for the vault (-v/--vault) allows to skip the
    creation of the vault if this is provided.
    """
    parser = ArgumentParser(usage=help_text)
    parser.add_argument("-v", "--vault", dest="vault", required=False, help="Path to the vault")
    parser.add_argument("-s", "--input-source", dest="input_source", required=True,
                        help="Path to the processed sources")
    parser.add_argument("-r", "--routine", dest="routine_name", required=True,
                        help="Name of the routine to create different interfaces")

    args = parser.parse_args()

    return args.vault, args.input_source, args.routine_name


# Few functions
def merge_declaration_lines(code_text):
    lines = [l for l in code_text.split("\n")]
    for _index, _line in enumerate(lines):
        line = remove_comments(_line).strip()
        if line.count("&") and line.count("::"):
            i = 1
            while line and line.strip()[-1] == "&":
                line1 = line
                line2 = remove_comments(lines[_index + i]).strip()
                line = line1.replace("&", "").rstrip() + line2
                i += 1
            line = line.replace("&", "")
            line = re.sub(", +", ", ", line)
            line = re.sub(":  +", ":  ", line)
            lines[_index] = line
            for j in range(1, i):
                lines[_index + j] = ""
    return "\n".join(lines).strip()


def replace_subroutine_name(old_name, new_name, code_text):
    pattern = r"(subroutine +\b)%s\b" % old_name
    pat = re.compile(pattern, re.I)
    replacement = r"\1 %s" % new_name
    return pat.sub(replacement, code_text)


def find_declaration_line(variable, routine_lines):
    import re
    import warnings
    # Declaration pattern
    # TODO: This might have issues with split lines
    pattern = r"::.*\b%s\b" % variable.name

    exception_pattern = r"&*\b%s\b" % variable.name
    real_pattern = r"real.*::"

    current_block = CurrentBlock(vault[variable.procedure.module.name].main())
    for _index, _line in enumerate(routine_lines):
        # _line = remove_comments(_line).strip()

        if re.search(pattern, _line):
            if current_block.procedure.name == variable.procedure.name:
                return _index
        else:
            current_block = current_block.in_which_block(_line, vault=vault)

    print("Declaration of variable %s with id %s not found in module %s" % (variable.name, variable.id, variable.procedure.module.name))
    return None


def obtain_routine_source(routine, sources_path):
    source_file = join(sources_path, "%s.f90" % routine.module.name)

    with open(source_file) as f:
        # Routine name pattern
        pattern = r"routine.*\b%s\b" % routine.name
        lines = [l for l in f]
        indices = []

        for index, line in enumerate(lines):
            line = remove_comments(line).strip()

            # Search for the start/end of the routine
            if re.search(pattern, line, re.I):
                indices.append(index)

        routine_code = "".join(lines[indices[0]:indices[1] + 1])

    routine_code = merge_declaration_lines(routine_code)
    return routine_code, indices


def replace_routine_name(old_name, new_name, code_text):
    pattern = r"(subroutine +\b)%s\b" % old_name
    pat = re.compile(pattern, re.I)
    replacement = r"\1 %s" % new_name
    return pat.sub(replacement, code_text)


if __name__ == "__main__":

    # Get command line arguments
    vault_path, path_to_input_sources, routine_name = get_command_line_arguments()

    path_to_preprocessed_sources = "./tmp_processedsources/"
    # In case no vault has been provided it will create it from the sources in the input sources folder
    if vault_path is None:
        # Paths to source folders
        # Path_to_input_sources
        if not isdir(path_to_preprocessed_sources):
            mkdir(path_to_preprocessed_sources)

        # Format text + replace real declarations with type(rpe_var)
        preprocess_sources(path_to_input_sources, path_to_preprocessed_sources)

        # Obtain source information
        vault = parse_sources(path_to_preprocessed_sources, save_vault=True)
    else:
        vault = load_vault(vault_path)

    routine = None
    for procedure in vault.procedures:
        if procedure.name == routine_name:
            routine = procedure
            break
    if routine is None:
        print("Routine %s not found in vault." % routine_name)
        exit(0)

    routine_code, indices = obtain_routine_source(routine, path_to_input_sources)

    routine_lines = routine_code.split("\n")
    sensitive_variables = [v for v in routine.variables if v.is_dummy_argument and v.type == "double"]

    # All possible combinations of precisions
    precision_combinations = list(itertools.product(["dp", "sp"], repeat=len(sensitive_variables)))

    # Get filepath
    filepath = join(path_to_input_sources, "%s.f90" % routine.module.name)

    code_versions = {}
    interface = "module procedure "
    for c_index, combination in enumerate(precision_combinations):
        version_lines = routine_lines[:]
        new_routine_name = "%s_%04i" % (routine.name, c_index)
        for index, variable in enumerate(sensitive_variables):
            variable_precision = combination[index]
            dec_line = find_declaration_line(variable, routine_lines)
            original_line = version_lines[dec_line]
            new_line = replace_real_declaration(variable, original_line, variable_precision)
            version_lines[dec_line] = new_line
        version_text = "\n".join(version_lines).strip()
        version_text = replace_routine_name(routine.name, new_routine_name, version_text)
        code_versions[c_index] = version_text
        if c_index == 0:
            interface += "%s" % new_routine_name
        else:
            interface += ", %s" % new_routine_name

    interface = "interface %s\n           %s\nend interface %s" % (routine.name, interface, routine.name)
    for key, code in code_versions.items():
        print(code)
    # Deleting temporary folders
    if isdir(path_to_preprocessed_sources):
        rmtree(path_to_preprocessed_sources)
