import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
from AutoRPE.UtilsRPE.BasicFunctions import close_brackets
from AutoRPE.UtilsRPE.CallManager import find_call_arguments, find_called_function, strip_call_arguments
import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
import AutoRPE.UtilsRPE.Finder as Finder
import AutoRPE.UtilsRPE.Getter as Getter
from AutoRPE.UtilsRPE.AttributeParser import parse_attributes, get_variable_name
import AutoRPE.UtilsRPE.CurrentBlock as CurrentBlock
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import re


# File manipulation Classes
class ReadSourceFile(object):
    def __init__(self, _filename):
        # Init variables
        self.file_path = _filename
        self.filename = self.file_path.split("/")[-1]
        self.text = None
        self.lines = None

        # Read file
        self.read_file(_filename)

    def read_file(self, _filename):
        try:
            with open(_filename) as f:
                self.text = f.read()
                self.lines = self.text.split("\n")
        except UnicodeDecodeError:
            with open(_filename, encoding="latin1") as f:
                self.text = f.read()
                self.lines = self.text.split("\n")

    def rebuild_text(self):
        self.text = "".join([l + "\n" for l in self.lines[:]])

    def save_modified_file(self, filepath):
        with open(filepath, "w") as output_file:
            output_file.write(self.text)


class ObtainSourcefileInfo(ReadSourceFile):
    # Definition: This class is used to load a sourcefile and parse the subroutines contained there
    # Input: string with path to sourcefile
    def __init__(self, _filename):
        super(ObtainSourcefileInfo, self).__init__(_filename)
        # Definition: Initializing the object reading the sourcefile
        # Input: string with path to sourcefile
        # I think that it is safe to collect the information using all the code in lower case.
        self.text = self.text.lower()
        self.objects = []
        self.module = None

    # Definition: Given a line number, looks for next line with "end subroutine"
    # Input: integer with line number
    # Output: integer with the line number of the next "end subroutine"
    def find_next_subroutine_end(self, _index):
        return self.find_next_end(_index, "end subroutine")

    # Definition: Given a line number, looks for next line with "end function"
    # Input: integer with line number
    # Output: integer with the line number of the next "end function"
    def find_next_function_end(self, _index):
        return self.find_next_end(_index, "end function")

    # Definition: Given a line number, looks for next line with "end interface"
    # Input: integer with line number
    # Output: integer with the line number of the next "end interface"
    def find_next_interface_end(self, _index):
        return self.find_next_end(_index, "end interface")

    # Definition: Given a line number, looks for next line with the string contained in end_statement argument.
    # Input:
    #       _index: integer with line number
    #       end_statement: string that indicates next end_statement
    # Output: integer with the line number of the next line containing the end_statement
    def find_next_end(self, _index, end_statement):
        # for _line_index, _line in enumerate(self.lines[_index:]):
        #     if _line.lower().count(end_statement):
        #         return _line_index
        # return len(self.lines)
        _i = _index + 1
        begin_statement = end_statement.replace('end', '')
        nested_declaration = 0
        while _i < len(self.lines):
            _l = self.lines[_i]
            if re.match('^\s*(' + begin_statement + ')', _l, re.I):
                nested_declaration += 1
            if _l.lower().count(end_statement):
                if nested_declaration == 0:
                    return _i
                else:
                    nested_declaration -= 1
            _i += 1

        return len(self.lines)

    # Definition: Given a line number, returns the subroutine name and the dummy arguments (in case they exist)
    # Input: integer with line number
    # Output: tuple with routine name and list of dummy arguments
    def get_dummy_arguments(self, _index):
        _line = self.lines[_index]
        # Workaround to solve the problem that appears when there's a function declaration with (kind=...)
        if _line.lower().count("function "):
            if _line.find("(") < _line.lower().find("function "):
                _line = _line[_line.lower().find("function "):]

        # If the subroutine declaration does not have any parenthesis, this routine does not have any argument
        if not _line.count("("):
            _name = _line.split()[1].strip()
            _dummy_arguments = []
        # If otherwise it has dummy arguments, find them
        elif _line.lower().count("result") or _line.lower().count("bind("):
            _name = _line[:_line.find("(")].split()[-1]
            _dummy_arguments = []
            _dummy_arguments += strip_call_arguments(close_brackets(_line))
            # I don't know if we should record the result variable with the other variables
            # result_start = _line.lower().find("result")
            # _dummy_arguments += strip_call_arguments(_line[result_start:])
        else:
            _name = _line[:_line.find("(")].split()[-1]
            _dummy_arguments = strip_call_arguments(_line)
        _dummy_arguments = [_x for _x in _dummy_arguments if _x.strip()]
        return _name, _dummy_arguments

    # Definition: Given the line numbers where a routine starts and ends,
    # look for variable declarations.
    # Input:
    #       -_subprogram_start: integer with line number where the routine starts
    #       -_subprogram_end: integer with line number where the routine ends
    # Output: list of strings containing declarations
    def find_subprogram_declaration(self, _subprogram_start, _subprogram_end, obj_to_find="::"):
        # Check if there is some statement inside
        # TODO find a better way of parsing this: actually we can take advantage of the fact that
        # Cases in which a routine contains other functions declarations.
        stop_index = [index
                      for index, line in enumerate(self.lines[_subprogram_start:_subprogram_end + 1])
                      if line.count("contains") or line.count("interface")]
        # A declaration found stop the search there
        if stop_index:
            _subprogram_end = _subprogram_start - 1 + stop_index[0]

        # Lines inside the routine
        _subprogram_lines = self.lines[_subprogram_start:_subprogram_end + 1]

        # Lines that are variable declarations
        _all_declaration = [_l.strip() for _l in _subprogram_lines if
                            _l.count(obj_to_find) and not
                            _l.lower().count("allocate")]

        return _all_declaration

    # Definition: Given the routine name, the name of the dummy arguments and the lines containing declarations,
    # this method finds the attributes of each dummy arguments and adds all the information
    # into the object variable "routines".
    # Input:
    #       -routine_name: string with the routine name
    #       -dummy_arguments: list of strings with argument names
    #       -declarations: list of strings with variable declarations
    def add_struct_members(self, file, structure):
        try:
            for obj in file.objects:
                if obj.name == "main":
                    main = obj
                    break
            for member in structure.variables:
                main.add_variable(member)

        except UnboundLocalError:
            print("Main not found for structure %s", structure.name)

    def find_procedure_variables(self, header, routine_name, dummy_arguments, var_declarations, obj_type, used_modules):
        if obj_type == BasicStructures.Function:
            # Retrieve the ret_val name
            ret_type, ret_val = Finder.find_function_type_from_header(header, var_declarations)
            if not ret_val:
                ret_val = routine_name
        elif obj_type == BasicStructures.SubRoutine:
            ret_val = None
        else:
            raise ValueError("Header does not correspond to function or routine")

        ob = obj_type(routine_name, self.module)

        # ob.used_modules = [m.split("use")[1].split("only")[0].replace(",", "").strip() for m in used_modules]
        if len(used_modules):
            ob.use_only_variables = [m.split("use")[1].strip() for m in used_modules]
        # Reverse approach
        dummy_arguments = [x.lower() for x in dummy_arguments]
        for declaration in var_declarations:
            pattern = "(.*)::(.*)"
            m = re.search(pattern, declaration)
            # In this case is not a variable declaration
            if declaration.count("contiguous"):
                continue
            _attributes, variable = m.groups()
            # Check if the dummy argument is inside the declaration arguments
            # If it is, take the type, dimension, intent and optional
            _var_type, _dimension, _intent, _optional, \
            _allocatable, is_pointer, is_parameter = parse_attributes(_attributes, variable)

            variable_name = get_variable_name(variable)

            if len(variable_name) != 1:
                raise AssertionError("Problem identifying variable declaration.")
            variable_name = variable_name[0]
            _v = BasicStructures.Variable(ob, variable_name, _var_type, _dimension, self.module)
            _v.allocatable = _allocatable
            _v.is_parameter = is_parameter
            # If the variable name is the same of the function or is return statement
            if ret_val is not None and _v.name.lower() == ret_val.lower():
                _v.is_retval = True
                ob.type = _var_type
            _v.is_pointer = is_pointer
            if variable_name.lower() in dummy_arguments:
                _v.intent = _intent
                _v.is_optional = _optional
                _v.is_dummy_argument = True
            ob.add_variable(_v)

        # Save the position of the argument
        found_vars = [variable.name.lower() for variable in ob.variables]

        for index, dummy_argument in enumerate(dummy_arguments):
            if dummy_argument in found_vars:
                # TODO: save position of already found vars
                ob.get_variable_by_name(dummy_argument).position = index  # variable.position = index
                continue
                # If it's not found its an implicitly declared variable
            if dummy_argument[0] in "ijklmn":
                var_type = "integer"
            else:
                var_type = VariablePrecision.wp
            dimension = 0
            intent = None
            optional = None
            _v = BasicStructures.Variable(ob, dummy_argument, var_type, dimension, self.module)
            _v.intent = intent
            _v.is_optional = optional
            _v.is_dummy_argument = True
            _v.position = index
            _v.is_parameter = False
            ob.add_variable(_v)

        if obj_type == BasicStructures.Function:
            # In case of function  declaration like INTEGER FUNCTION name() without variables declaration
            if not [var for var in ob.variables if var.is_retval]:
                # Add the return value to functions variables
                _v = BasicStructures.Variable(ob, ret_val, ret_type, 0, self.module)
                _v.is_retval = True
                ob.add_variable(_v)
                # Specify the type of the function
                ob.type = ret_type
        # Save object
        self.objects.append(ob)

    # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
    # dimensions and optionality of the arguments.
    def find_routines(self):

        # Look into each line
        for _index, line in enumerate(self.lines):
            # Look if routine starts with (recoursive) subroutine
            if re.search("^\s*\w*\s*subroutine", line, re.IGNORECASE):
                if line.lower().strip().find("end") == 0:
                    continue
                subroutine_start = _index
                # Look for end subroutine statements
                subroutine_end = self.find_next_subroutine_end(_index)

                # Find name and dummy arguments
                routine_name, dummy_arguments = self.get_dummy_arguments(_index)
                # Find declarations
                var_declarations = self.find_subprogram_declaration(subroutine_start, subroutine_end, obj_to_find="::")
                used_modules = self.find_subprogram_declaration(subroutine_start, subroutine_end, obj_to_find="use ")
                header = self.lines[subroutine_start].lower()
                # Parse everything and store
                self.find_procedure_variables(header, routine_name.lower(), dummy_arguments, var_declarations,
                                              obj_type=BasicStructures.SubRoutine, used_modules=used_modules)

    # ERROR: Definition: Looks into the sourcefile for routine declarations, storing the name, type,
    # dimensions and optionality of the arguments.
    def find_types(self):
        import AutoRPE.UtilsRPE.RegexPattern as RegexPatter
        # Look into each line
        in_type_declaration = False
        types = {}
        type_name = None
        variable_counter = None
        for _index, line in enumerate(self.lines):
            # Look if routine starts with subroutine
            if not in_type_declaration:
                # check if a type has been declared
                m = re.search(RegexPatter.simple_type_declaration, line, re.IGNORECASE)
                if m:
                    in_type_declaration = True
                    variable_counter = 0
                    type_name = m.groups(0)[0]
                    types[type_name] = {}
                    ob = BasicStructures.DerivedType(type_name, self.module)
                    self.objects.append(ob)
                # Case in which a type is extending another type
                match_extend = re.search(RegexPatter.extended_type_declaration, line, re.IGNORECASE)
                if match_extend:
                    in_type_declaration = True
                    variable_counter = 0
                    type_name = match_extend.groups()[2]
                    extended_type = match_extend.groups()[1]
                    types[type_name] = {}
                    ob = BasicStructures.DerivedType(type_name, self.module, extended_type)
                    self.objects.append(ob)

            else:
                # Check if we are exiting a type declaration
                m = re.search(RegexPatter.end_type_statement, line, re.IGNORECASE)
                if m:
                    in_type_declaration = False
                else:
                    pattern = "(.*)::(.*)"
                    m = re.search(pattern, line)
                    if m:
                        attributes, variables = m.groups()
                        var_type, dimension, intent, optional, \
                        allocatable, is_pointer, is_parameter = parse_attributes(attributes, variables)

                        variables = get_variable_name(variables)
                        for var in variables:
                            types[type_name][var] = var_type
                            v = BasicStructures.Variable(ob, var, var_type, dimension, self.module)
                            v.position = variable_counter
                            v.is_dummy_argument = True
                            v.allocatable = allocatable
                            v.is_parameter = is_parameter
                            v.is_pointer = is_pointer
                            v.is_member_of = type_name
                            ob.add_variable(v)
                            variable_counter += 1
        return types

    def find_module_variables(self):
        import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
        main_routine = BasicStructures.ModuleMain(self.filename)

        line_to_check = True
        self.objects.append(main_routine)
        pattern = "(.*)::(.*)"
        for line in self.lines:
            if line_to_check:
                # Check if we are entering a type declaration zone
                is_type_declaration = re.search(RegexPattern.general_type_declaration, line, re.IGNORECASE)
                if is_type_declaration or NatureDeterminer.is_start_function_statement(line):
                    line_to_check = False
                    continue
            else:
                # Check if we are exiting a type declaration zone
                is_end_type_declaration = re.search(RegexPattern.end_type_statement, line, re.IGNORECASE)
                if is_end_type_declaration or NatureDeterminer.is_end_function_statement(line):
                    line_to_check = True
                # Even if we are inside or exiting a type declaration zone, we have to jump to next line
                continue

            if line.lower().count("contains"):
                break
            if line.lower().count("end program"):
                break
            match = re.search(pattern, line)
            if match:
                attributes, variables = match.groups()
                var_type, dimension, intent, optional, \
                allocatable, is_pointer, is_parameter = parse_attributes(attributes, variables)
                variables = get_variable_name(variables)
                if var_type:
                    for variable in variables:
                        v = BasicStructures.Variable(main_routine, variable, var_type, dimension, self.module)
                        v.allocatable = allocatable
                        v.is_pointer = is_pointer
                        v.is_parameter = is_parameter
                        main_routine.add_variable(v)

    # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
    # dimensions and optionality of the arguments.
    def find_functions(self):
        # Look into each line
        for _index, line in enumerate(self.lines):
            if NatureDeterminer.is_start_function_statement(line):
                function_start = _index
                # Look for end subroutine statements
                function_end = self.find_next_function_end(_index)

                # Find name and dummy arguments
                function_name, dummy_arguments = self.get_dummy_arguments(_index)
                # Find declarations
                variables_declarations = self.find_subprogram_declaration(function_start, function_end)
                used_modules = self.find_subprogram_declaration(function_start, function_end, obj_to_find="use")
                # Parse everything and store
                header = self.lines[function_start].lower()
                self.find_procedure_variables(header, function_name.lower(), dummy_arguments, variables_declarations,
                                              obj_type=BasicStructures.Function, used_modules=used_modules)

    def find_implicit_functions(self, vault):
        try:
            current_block = CurrentBlock.CurrentBlock(self.module.main())
        except AttributeError:
            return None
        # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
        # dimensions and optionality of the arguments.
        pattern = "^ *(\w+) *\([\w, ]*\) *= *(.*)"
        # Look into each line
        for _index, line in enumerate(self.lines):
            m = re.search(pattern, line)
            if m:
                procedure = current_block.procedure
                call = close_brackets(line[m.start(1):])
                call_arguments = find_call_arguments(call)
                is_function = False
                for argument in call_arguments:
                    argument_type = Getter.get_type_of_contents(argument, procedure=procedure, vault=vault)
                    if argument_type != "integer":
                        is_function = True
                        break

                if is_function:
                    function_name = find_called_function(call)
                    new_function = BasicStructures.Function(function_name, module=self.module)
                    new_function.type = Getter.get_type_of_contents(function_name, procedure, vault=vault)
                    for argument_index, argument in enumerate(call_arguments):
                        var = procedure.get_variable_by_name(argument)
                        var.is_dummy_argument = True
                        var.position = argument_index
                        var.intent = "in"
                        new_function.add_variable(var)
                    self.objects.append(new_function)
            else:
                current_block = current_block.in_which_block(line, vault=vault)

    # Definition: Looks into the sourcefile for interface declarations, storing the name.
    def find_interfaces(self, object_procedures):
        for _index, _line in enumerate(self.lines):
            if _line.strip().lower().find("interface") == 0:
                interface_start = _index
                interface_end = self.find_next_interface_end(_index)
                try:
                    interface_name = _line.split()[1].strip()
                except IndexError:
                    # interface to an external procedure or a procedure used in a procedure with the DEFERRED attribute
                    continue
                procedure = []
                line_procedure = None
                for interface_line in self.lines[interface_start:interface_end + 1]:
                    if interface_line.strip().lower().find("module procedure") == 0:
                        line_procedure = interface_line.lower().replace("module procedure", "")
                        line_procedure = [_r.strip() for _r in line_procedure.split(",")]
                        procedure += line_procedure
                    elif interface_line.strip().lower().strip().find("subroutine") == 0:
                        subroutine_name = interface_line[:interface_line.find("(")].split()[1]
                        procedure += [subroutine_name]

                interface_procedures = [object_procedures[p] for p in procedure]
                ob = BasicStructures.Interface(interface_name)
                for procedure in interface_procedures:
                    ob.add_procedure(procedure)
                if not ob.subprogram:
                    raise AssertionError("No routine found for interface %s" % interface_name)
                self.objects.append(ob)

    def find_used_modules(self, vault):
        current_block = CurrentBlock.CurrentBlock(self.module.main())
        for line in self.lines[1:]:
            current_block = current_block.in_which_block(line, vault=vault)
            line = line.strip().lower()
            # No need to parse the entire file
            if not line.find("use ") == 0:
                break
            module_name = line.split()[1]
            if module_name.count(","):
                module_name = module_name.replace(",", "")
            for module in vault.modules:
                if module.name == module_name:
                    # Sometimes in Nemo use statement are repeated
                    if not len([m.name for m in self.module.uses if m.name == module_name]) :
                        self.module.uses.append(module)

            # Look for
            pattern = "(\w+) *=> *(\w+)"
            matches = re.findall(pattern, line)
            if matches:
                for match in matches:
                    pointer, target = match
                    target_var = self.module.uses[-1].main().get_variable_by_name(target)
                    new_var = BasicStructures.Variable(current_block.procedure, pointer, target_var.type, target_var.dimension)
                    current_block.procedure.add_variable(new_var)
