# This file contains the definition of a Binary Search Job, which its an extension of a Job class.
# The extension includes several methods that are useful to do a precision analysis using a binary search algorithm.
#
# Import the Job class
from AutoRPE.UtilsWorkflow.Job import Job


# Extend it
class BinarySearchJob(Job):
    def __init__(self,
                 analysis_variables,
                 forced_variables=None,
                 reduced_precision_level=10,
                 communicator=None,
                 vault=None,
                 template="",
                 local_folder=None,
                 accuracy_test=None,
                 ):
        super().__init__(analysis_variables,
                         forced_variables=forced_variables,
                         reduced_precision_level=reduced_precision_level,
                         communicator=communicator,
                         vault=vault,
                         template=template,
                         result_filename=accuracy_test.result_filename)
        # This class extends the classical job putting some extra variables that define the relations
        self.parent = None
        self.child = None
        self.child_number = None
        self.level = None

        # Other thinks that need to be saved
        self.accuracy_test = accuracy_test
        self.banned_variables = []
        self.local_folder = local_folder
        self.max_retrials = 1
        self.retrials = 0

    def identifier(self):
        from AutoRPE.UtilsWorkflow.Utils import var_set_identifier
        return var_set_identifier(self)

    # The following functions have been extracted from this code and are now provided as an Evaluator class, which
    # is stored in self.accuracy_test, and which has the evaluate_success, evaluate_fitness and plot methods that are
    # used here.
    def evaluate_success(self):
        return self.accuracy_test.evaluate_success(self.result())

    def evaluate_fitness(self):
        return self.accuracy_test.evaluate_fitness(self.result())

    def create_plots(self):
        return self.accuracy_test.plot(self.result())

    def ancestors(self):
        to_return = []
        cv = self
        while cv.parent is not None:
            to_return.append(cv.parent)
            cv = cv.parent
        return list(to_return)

    def descendants(self):
        # Using a recursive approach to obtain the full list of descendants
        if self.child is None:
            return []
        else:
            # Getting the two lists of descendants of the two child and merging them in a single list
            child_descendants = sum([child.descendants() for child in self.child], [])
            # Returning the child plus their descendants
            return self.child[:] + child_descendants

    def spawn_children(self):
        from AutoRPE.UtilsWorkflow.Utils import divide_set
        # In case the set has not been divided already
        if self.child is None:

            # Trying new thing, set the minimum unit to explore to modules
            # Check if all variables are from the same module
            variables = [self.vault.get_variable_by_id(_id) for _id in self.analysis_set]
            routines = set([v.procedure for v in variables])
            modules = set([r.module for r in routines])
            if len(modules) == 1:
                return False
            if len(self.analysis_set) >= 2:
                _set1, _set2 = divide_set(self.analysis_set, self.vault)
                self.child = []
                for index, var_set in enumerate([_set1, _set2]):
                    job = BinarySearchJob(var_set,
                                          forced_variables=self.forced_variables,
                                          reduced_precision_level=self.reduced_precision_level,
                                          communicator=self.communicator,
                                          vault=self.vault,
                                          template=self.template,
                                          accuracy_test=self.accuracy_test,
                                          local_folder=self.local_folder,
                                          )
                    job.parent = self
                    job.child_number = index
                    job.level = self.level + 1
                    self.child.append(job)
                return True
            else:
                return False
        else:
            return False

    def driver(self, queue, max_running_jobs):
        status = self.status
        if self.status == "PENDING":
            if len(queue.running) < max_running_jobs:
                self.submit()
                self.status = "RUNNING"
                queue.pending.remove(self)
                queue.running.append(self)
        elif self.status == "RUNNING":
            self.check_running(queue)
        elif self.status == "SUSPENDED":
            self.check_child(queue)

        # Return True if there was a status change, False otherwise
        return status != self.status

    def check_running(self, queue):
        remote_status = self.remote_status
        if remote_status == "COMPLETED":
            self.retrials = 0
            success = self.evaluate_success()
            # Create plots
            # self.create_plots()
            if success:
                self.qualify(queue)
            else:
                self.fail(queue)

        elif remote_status == "FAILED":
            if self.retrials < self.max_retrials:
                self.submit(force=True)
                self.retrials += 1
            else:
                raise AssertionError("SLURM job failed!")
        elif remote_status == 'CANCELLED+':
            self.submit(force=True)
            self.retrials = 0
        elif remote_status == "RUNNING":
            pass
        elif remote_status == "PENDING":
            pass
        elif remote_status == "TIMEOUT":
            self.submit(force=True)
        else:
            raise AssertionError("Unhandled status %s" % remote_status)

    def move_to_success(self, queue):
        queue[self.status].remove(self)
        self.status = "SUCCESS"
        queue.success.append(self)

    def qualify(self, queue):
        self.move_to_success(queue)

    def fail(self, queue):
        from AutoRPE.UtilsWorkflow.Errors import NoException, IntraRoutine, InterModule, IntraModule
        # If the set can be split
        if self.spawn_children():
            print("Job %s %s suspended with fitness: %.2f." % (self.identifier(), self.hash, self.evaluate_fitness()))
            queue[self.status].remove(self)
            queue.suspended.append(self)
            self.status = "SUSPENDED"
            for ch in self.child:
                if ch not in queue.all():
                    queue.pending.append(ch)
        # If the set can not be split, check whether its a exception and handle it
        else:
            if self.kind_of_exception() == NoException:
                queue[self.status].remove(self)
                self.status = "FAILED"
                queue.failed.append(self)
                print("Job %s %s failed with fitness: %.2f." % (self.identifier(), self.hash, self.evaluate_fitness()))
            else:
                # Exception handling
                print(
                    "Job %s %s suspended with fitness: %.2f." % (self.identifier(), self.hash, self.evaluate_fitness()))
                print("Exception found!", self.kind_of_exception())
                queue[self.status].remove(self)
                queue.suspended.append(self)
                self.status = "SUSPENDED"

                child_1, child_2 = self.child
                # Make that the analysis variables of one child are the forced variables of the other one
                if child_1.forced_variables is not None:
                    child_1.forced_variables.extend(child_2.analysis_set)
                    child_2.forced_variables.extend(child_1.analysis_set)
                else:
                    child_1.forced_variables = child_2.analysis_set
                    child_2.forced_variables = child_1.analysis_set
                for ch in self.child:
                    if ch in queue.all():
                        queue[ch.status].remove(ch)
                    queue.pending.append(ch)
                    ch.status = "PENDING"
                    # Disinherit all child's descendants
                    for job in ch.descendants():
                        job.parent = None
                    ch.child = None
                # Save disinherited jobs
                disinherited = [j for j in queue.all() if j.parent is None and j.level != 0]
                for job in disinherited:
                    queue[job.status].remove(job)
                    if job not in queue.disinherited:
                        queue.disinherited.append(job)

    def check_child(self, queue):
        # Check if all the child jobs have finished
        all_finished = all([c.status == "FAILED" or c.status == "SUCCESS" for c in self.child])
        if all_finished:
            for child in self.child:
                # Get the banned variables from all the child
                self.banned_variables += child.banned_variables

                # In case the child job failed, include its analysis variables in the banned variables
                if child.status == "FAILED":
                    self.banned_variables += child.analysis_set
                    # Keep unique values only
                    self.banned_variables = list(set(self.banned_variables))
            self.analysis_set = [var for var in self.analysis_set if var not in self.banned_variables]

            if self.analysis_set is []:
                self.fail(queue)
            else:
                self.status = "PENDING"
                queue.pending.append(self)
                queue.suspended.remove(self)

    def kind_of_exception(self):
        """
        This method tries to categorize the kind of exception that occurs when two successful sets merge to an
        unsuccessful one.

        To start with something I propose:
        - No exception: It isn't a failed job
        - Intra-routine: All variables belong to same routine
        - Intra-module: All variables belong to same module, but not routine
        - Inter-module: The variables belong to different modules.
        :return: kind-of-exception
        """
        from AutoRPE.UtilsWorkflow.Errors import NoException, IntraRoutine, InterModule, IntraModule
        if not self.child:
            return NoException
        child_1, child_2 = self.child
        # We'll focus only on the analysis_set
        ch1_variables = [self.vault.get_variable_by_id(_id) for _id in child_1.analysis_set]
        ch2_variables = [self.vault.get_variable_by_id(_id) for _id in child_2.analysis_set]
        ch1_routines = set([v.procedure for v in ch1_variables])
        ch2_routines = set([v.procedure for v in ch2_variables])
        if len(ch1_routines.union(ch2_routines)) == 1:
            return IntraRoutine
        ch1_modules = set([v.procedure.module for v in ch1_variables])
        ch2_modules = set([v.procedure.module for v in ch2_variables])
        if len(ch1_modules.union(ch2_modules)) == 1:
            return IntraModule
        return InterModule